@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -51,6 +51,7 @@ static void
 set_http_error (http_status_t status)
 {
   PyObject *v = Py_BuildValue ("i", status);
+  debugprintf("set_http_error: %d\n", (int) status);
   if (v != NULL) {
     PyErr_SetObject (HTTPError, v);
     Py_DECREF (v);
@@ -64,6 +65,7 @@ set_ipp_error (ipp_status_t status)
 
   last_error = ippErrorString (status);
 
+  debugprintf("set_ipp_error: %d, %s\n", (int) status, last_error);
   PyObject *v = Py_BuildValue ("(is)", status, last_error);
   if (v != NULL) {
     PyErr_SetObject (IPPError, v);
@@ -97,6 +99,7 @@ Connection_init (Connection *self, PyObject *args, PyObject *kwds)
   debugprintf ("-> Connection_init()\n");
 
   Py_BEGIN_ALLOW_THREADS;
+  debugprintf ("httpConnectEncrypt(...)\n");
   self->http = httpConnectEncrypt (cupsServer (),
 				   ippPort (),
 				   cupsEncryption ());
@@ -115,8 +118,10 @@ Connection_init (Connection *self, PyObject *args, PyObject *kwds)
 static void
 Connection_dealloc (Connection *self)
 {
-  if (self->http)
+  if (self->http) {  
+    debugprintf ("httpClose()\n");
     httpClose (self->http);
+  }
 
   self->ob_type->tp_free ((PyObject *) self);
 }
@@ -126,37 +131,65 @@ Connection_dealloc (Connection *self)
 ////////////////
 
 static PyObject *
-do_printer_request (Connection *self, PyObject *args, ipp_op_t op)
+do_printer_request (Connection *self, PyObject *args, PyObject *kwds,
+		    ipp_op_t op)
 {
   const char *name;
+  const char *reason = NULL;
   char uri[HTTP_MAX_URI];
   ipp_t *request, *answer;
 
-  if (!PyArg_ParseTuple (args, "s", &name))
-    return NULL;
+  switch (op) {
+  case IPP_PAUSE_PRINTER:
+  case CUPS_REJECT_JOBS:
+    {
+      static char *kwlist[] = { "name", "reason", NULL };
+      if (!PyArg_ParseTupleAndKeywords (args, kwds, "s|s", kwlist,
+					&name, &reason))
+	return NULL;
+      debugprintf ("-> do_printer_request(op:%d, name:%s, reason:%s)\n",
+		   (int) op, name, reason ? reason : "(null)");
+      break;
+    }
+
+  default:
+    if (!PyArg_ParseTuple (args, "s", &name))
+      return NULL;
+    debugprintf ("-> do_printer_request(op:%d, name:%s)\n", (int) op, name);
+    break;
+  }
 
   request = ippNewRequest (op);
   snprintf (uri, sizeof (uri), "ipp://localhost/printers/%s", name);
   ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
 		"printer-uri", NULL, uri);
+
+  if (reason)
+    ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_TEXT,
+		  "printer-state-message", NULL, reason);
+
+  debugprintf ("cupsDoRequest(\"/admin/\")\n");
   answer = cupsDoRequest (self->http, request, "/admin/");
   if (PyErr_Occurred ()) {
     if (answer)
       ippDelete (answer);
+    debugprintf("<- do_printer_request (error)\n");
     return NULL;
-  
-}
+  }
+
   if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
     set_ipp_error (answer ?
 		   answer->request.status.status_code :
 		   cupsLastError ());
     if (answer)
       ippDelete (answer);
+    debugprintf("<- do_printer_request (error)\n");
     return NULL;
   }
 
   ippDelete (answer);
   Py_INCREF (Py_None);
+  debugprintf("<- do_printer_request (None)\n");
   return Py_None;
 }
 
@@ -164,10 +197,14 @@ static PyObject *
 Connection_getDests (Connection *self)
 {
   cups_dest_t *dests;
-  int num_dests = cupsGetDests2 (self->http, &dests);
+  int num_dests;
   PyObject *pydests = PyDict_New ();
   int i;
 
+  debugprintf ("-> Connection_getDests()\n");
+  debugprintf ("cupsGetDests2()\n");
+  num_dests = cupsGetDests2 (self->http, &dests);
+
   // Create a dict indexed by (name,instance)
   for (i = 0; i < num_dests; i++) {
     PyObject *largs = Py_BuildValue ("()");
@@ -196,7 +233,9 @@ Connection_getDests (Connection *self)
     PyDict_SetItem (pydests, nameinstance, (PyObject *) destobj);
   }
 
+  debugprintf ("cupsFreeDests()\n");
   cupsFreeDests (num_dests, dests);
+  debugprintf ("<- Connection_getDests()\n");
   return pydests;
 }
 
@@ -213,19 +252,25 @@ Connection_getPrinters (Connection *self)
     "printer-info",
     "printer-make-and-model",
     "printer-state",
+    "printer-state-message",
+    "printer-state-reasons",
     "printer-uri-supported",
     "device-uri",
     "printer-is-shared",
   };
 
+  debugprintf ("-> Connection_getPrinters()\n");
+
   ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
 		 "requested-attributes",
 		 sizeof (attributes) / sizeof (attributes[0]),
 		 NULL, attributes);
+  debugprintf ("cupsDoRequest(\"/\")\n");
   answer = cupsDoRequest (self->http, request, "/");
   if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
     if (answer && answer->request.status.status_code == IPP_NOT_FOUND) {
       // No printers.
+      debugprintf ("<- Connection_getPrinters() = {} (no printers)\n");
       ippDelete (answer);
       return PyDict_New ();
     }
@@ -235,6 +280,7 @@ Connection_getPrinters (Connection *self)
 		   cupsLastError ());
     if (answer)
       ippDelete (answer);
+    debugprintf ("<- Connection_getPrinters() (error)\n");
     return NULL;
   }
 
@@ -254,6 +300,7 @@ Connection_getPrinters (Connection *self)
 	 attr = attr->next) {
       PyObject *val = NULL;
 
+      debugprintf ("Attribute: %s\n", attr->name);
       if (!strcmp (attr->name, "printer-name") &&
 	  attr->value_tag == IPP_TAG_NAME)
 	printer = attr->values[0].string.text;
@@ -301,6 +348,7 @@ Connection_getPrinters (Connection *self)
       }
 
       if (val) {
+	debugprintf ("Added %s to dict\n", attr->name);
 	PyDict_SetItemString (dict, attr->name, val);
 	Py_DECREF (val);
       }
@@ -316,6 +364,7 @@ Connection_getPrinters (Connection *self)
   }
 
   ippDelete (answer);
+  debugprintf ("<- Connection_getPrinters() = dict\n");
   return result;
 }
 
@@ -324,10 +373,13 @@ build_list_from_attribute_strings (ipp_attribute_t *attr)
 {
   PyObject *list = PyList_New (0);
   int i;
+  debugprintf ("-> build_list_from_attribute_strings()\n");
   for (i = 0; i < attr->num_values; i++) {
     PyObject *val = PyString_FromString (attr->values[i].string.text);
     PyList_Append (list, val);
+    debugprintf ("%s\n", attr->values[i].string.text);
   }
+  debugprintf ("<- build_list_from_attribute_strings()\n");
   return list;
 }
 
@@ -342,14 +394,17 @@ Connection_getClasses (Connection *self)
     "member-names",
   };
 
+  debugprintf ("-> Connection_getClasses()");
   ippAddStrings (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD,
 		 "requested-attributes",
 		 sizeof (attributes) / sizeof (attributes[0]),
 		 NULL, attributes);
+  debugprintf ("cupsDoRequest(\"/\")\n");
   answer = cupsDoRequest (self->http, request, "/");
   if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
     if (answer && answer->request.status.status_code == IPP_NOT_FOUND) {
       // No classes.
+      debugprintf ("<- Connection_getClasses() = {} (no classes)\n");
       ippDelete (answer);
       return PyDict_New ();
     }
@@ -359,6 +414,7 @@ Connection_getClasses (Connection *self)
 		   cupsLastError ());
     if (answer)
       ippDelete (answer);
+    debugprintf ("<- Connection_getClasses() (error)\n");
     return NULL;
   }
 
@@ -376,6 +432,7 @@ Connection_getClasses (Connection *self)
 
      for (; attr && attr->group_tag == IPP_TAG_PRINTER;
 	 attr = attr->next) {
+      debugprintf ("Attribute: %s\n", attr->name);
       if (!strcmp (attr->name, "printer-name") &&
 	  attr->value_tag == IPP_TAG_NAME)
 	classname = attr->values[0].string.text;
@@ -398,6 +455,7 @@ Connection_getClasses (Connection *self)
       members = PyList_New (0);
 
     if (classname) {
+      debugprintf ("Added class %s\n", classname);
       PyDict_SetItemString (result, classname, members);
     } else
       Py_XDECREF (members);
@@ -407,6 +465,7 @@ Connection_getClasses (Connection *self)
   }
 
   ippDelete (answer);
+  debugprintf ("<- Connection_getClasses() = dict\n");
   return result;
 }
 
@@ -417,6 +476,8 @@ Connection_getPPDs (Connection *self)
   ipp_t *request = ippNewRequest(CUPS_GET_PPDS), *answer;
   ipp_attribute_t *attr;
 
+  debugprintf ("-> Connection_getPPDs()\n");
+  debugprintf ("cupsDoRequest(\"/\")\n");
   answer = cupsDoRequest (self->http, request, "/");
   if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
     set_ipp_error (answer ?
@@ -424,6 +485,7 @@ Connection_getPPDs (Connection *self)
 		   cupsLastError ());
     if (answer)
       ippDelete (answer);
+    debugprintf ("<- Connection_getPPDs() (error)\n");
     return NULL;
   }
 
@@ -443,6 +505,7 @@ Connection_getPPDs (Connection *self)
 	 attr = attr->next) {
       PyObject *val = NULL;
 
+      debugprintf ("Attribute: %s\n", attr->name);
       if (!strcmp (attr->name, "ppd-name") &&
 	  attr->value_tag == IPP_TAG_NAME)
 	ppdname = attr->values[0].string.text;
@@ -476,21 +539,26 @@ Connection_getPPDs (Connection *self)
       }
 
       if (val) {
+	debugprintf ("Adding %s to ppd dict\n", attr->name);
 	PyDict_SetItemString (dict, attr->name, val);
 	Py_DECREF (val);
       }
     }
 
     if (ppdname) {
+      debugprintf ("Adding %s to result dict\n", ppdname);
       PyDict_SetItemString (result, ppdname, dict);
-    } else
+    } else {
+      debugprintf ("Discarding ppd dict\n");
       Py_DECREF (dict);
+    }
 
     if (!attr)
       break;
   }
 
   ippDelete (answer);
+  debugprintf ("<- Connection_getPPDs() = dict\n");
   return result;
 }
 
@@ -501,6 +569,8 @@ Connection_getDevices (Connection *self)
   ipp_t *request = ippNewRequest(CUPS_GET_DEVICES), *answer;
   ipp_attribute_t *attr;
 
+  debugprintf ("-> Connection_getDevices()\n");
+  debugprintf ("cupsDoRequest(\"/\")\n");
   answer = cupsDoRequest (self->http, request, "/");
   if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
     set_ipp_error (answer ?
@@ -508,6 +578,7 @@ Connection_getDevices (Connection *self)
 		   cupsLastError ());
     if (answer)
       ippDelete (answer);
+    debugprintf ("<- Connection_getDevices() (error)\n");
     return NULL;
   }
 
@@ -527,6 +598,7 @@ Connection_getDevices (Connection *self)
 	 attr = attr->next) {
       PyObject *val = NULL;
 
+      debugprintf ("Attribute: %s\n", attr->name);
       if (!strcmp (attr->name, "device-uri") &&
 	  attr->value_tag == IPP_TAG_URI)
 	device_uri = attr->values[0].string.text;
@@ -541,24 +613,237 @@ Connection_getDevices (Connection *self)
 	val = PyString_FromString (attr->values[0].string.text);
 
       if (val) {
+	debugprintf ("Adding %s to device dict\n", attr->name);
 	PyDict_SetItemString (dict, attr->name, val);
 	Py_DECREF (val);
       }
     }
 
     if (device_uri) {
+      debugprintf ("Adding %s to result dict\n", device_uri);
       PyDict_SetItemString (result, device_uri, dict);
-    } else
+    } else {
+      debugprintf ("Discarding device dict\n");
       Py_DECREF (dict);
+    }
 
     if (!attr)
       break;
   }
 
   ippDelete (answer);
+  debugprintf ("<- Connection_getDevices() = dict\n");
   return result;
 }
 
+static PyObject *
+Connection_getJobs (Connection *self, PyObject *args, PyObject *kwds)
+{
+  PyObject *result;
+  ipp_t *request = ippNewRequest(IPP_GET_JOBS), *answer;
+  ipp_attribute_t *attr;
+  char *which = NULL;
+  int my_jobs = 0;
+  static char *kwlist[] = { "which_jobs", "my_jobs", NULL };
+  if (!PyArg_ParseTupleAndKeywords (args, kwds, "|si", kwlist,
+				    &which, &my_jobs))
+	  return NULL;
+
+  debugprintf ("-> Connection_getJobs(%s,%d)\n",
+	       which ? which : "(null)", my_jobs);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI, "job-uri",
+		NULL, "ipp://localhost/jobs/");
+
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_KEYWORD, "which-jobs",
+		NULL, which ? which : "not-completed");
+
+  ippAddBoolean (request, IPP_TAG_OPERATION, "my-jobs", my_jobs);
+  if (my_jobs)
+    ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+		  "requesting-user-name", NULL, cupsUser());
+
+  debugprintf ("cupsDoRequest(\"/\")\n");
+  answer = cupsDoRequest (self->http, request, "/");
+  if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
+    set_ipp_error (answer ?
+		   answer->request.status.status_code :
+		   cupsLastError ());
+    if (answer)
+      ippDelete (answer);
+    debugprintf ("<- Connection_getJobs() (error)\n");
+    return NULL;
+  }
+
+  result = PyDict_New ();
+  for (attr = answer->attrs; attr; attr = attr->next) {
+    PyObject *dict;
+    int job_id = -1;
+
+    while (attr && attr->group_tag != IPP_TAG_JOB)
+      attr = attr->next;
+
+    if (!attr)
+      break;
+
+    dict = PyDict_New ();
+    for (; attr && attr->group_tag == IPP_TAG_JOB;
+	 attr = attr->next) {
+      PyObject *val = NULL;
+
+      debugprintf ("Attribute: %s\n", attr->name);
+      if (!strcmp (attr->name, "job-id") &&
+	  attr->value_tag == IPP_TAG_INTEGER)
+	job_id = attr->values[0].integer;
+      else if (((!strcmp (attr->name, "job-k-octets") ||
+		 !strcmp (attr->name, "job-priority") ||
+		 !strcmp (attr->name, "time-at-creation") ||
+		 !strcmp (attr->name, "time-at-processing") ||
+		 !strcmp (attr->name, "time-at-completed") ||
+		 !strcmp (attr->name, "job-media-sheets") ||
+		 !strcmp (attr->name, "job-media-sheets-completed")) &&
+		attr->value_tag == IPP_TAG_INTEGER) ||
+	       (!strcmp (attr->name, "job-state") &&
+		attr->value_tag == IPP_TAG_ENUM))
+	val = PyInt_FromLong (attr->values[0].integer);
+      else if ((!strcmp (attr->name, "job-name") &&
+		attr->value_tag == IPP_TAG_NAME) ||
+	       (!strcmp (attr->name, "job-originating-user-name") &&
+		attr->value_tag == IPP_TAG_NAME) ||
+	       (!strcmp (attr->name, "job-printer-uri") &&
+		attr->value_tag == IPP_TAG_URI))
+	val = PyString_FromString (attr->values[0].string.text);
+      else if (!strcmp (attr->name, "job-preserved") &&
+	       attr->value_tag == IPP_TAG_BOOLEAN)
+	val = PyBool_FromLong (attr->values[0].integer);
+
+      if (val) {
+	debugprintf ("Adding %s to job dict\n", attr->name);
+	PyDict_SetItemString (dict, attr->name, val);
+	Py_DECREF (val);
+      }
+    }
+
+    if (job_id != -1) {
+      debugprintf ("Adding %d to result dict\n", job_id);
+      PyObject *job_obj = PyInt_FromLong (job_id);
+      PyDict_SetItem (result, job_obj, dict);
+      Py_DECREF (job_obj);
+    }
+
+    Py_DECREF (dict);
+
+    if (!attr)
+      break;
+  }
+
+  ippDelete (answer);
+  debugprintf ("<- Connection_getJobs() = dict\n");
+  return result;
+}
+
+static PyObject *
+Connection_cancelJob (Connection *self, PyObject *args)
+{
+  ipp_t *request, *answer;
+  int job_id;
+  char uri[1024];
+  if (!PyArg_ParseTuple (args, "i", &job_id))
+    return NULL;
+
+  debugprintf ("-> Connection_cancelJob(%d)\n", job_id);
+  request = ippNewRequest(IPP_CANCEL_JOB);
+  snprintf (uri, sizeof (uri), "ipp://localhost/jobs/%d", job_id);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI, "job-uri", NULL, uri);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+		"requesting-user-name", NULL, cupsUser ());
+  debugprintf ("cupsDoRequest(\"/jobs/\")\n");
+  answer = cupsDoRequest (self->http, request, "/jobs/");
+  if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
+    set_ipp_error (answer ?
+		   answer->request.status.status_code :
+		   cupsLastError ());
+    if (answer)
+      ippDelete (answer);
+    debugprintf ("<- Connection_cancelJob() (error)\n");
+    return NULL;
+  }
+
+  Py_INCREF (Py_None);
+  debugprintf ("<- Connection_cancelJob() = None\n");
+  return Py_None;
+}
+
+static PyObject *
+Connection_setJobHoldUntil (Connection *self, PyObject *args)
+{
+  ipp_t *request, *answer;
+  int job_id;
+  char *job_hold_until = NULL;
+  char uri[1024];
+  cups_option_t *options = NULL;
+  int num_options = 0;
+  if (!PyArg_ParseTuple (args, "is", &job_id, &job_hold_until))
+    return NULL;
+
+  debugprintf ("-> Connection_setJobHoldUntil(%d,%s)\n",
+	       job_id, job_hold_until);
+  request = ippNewRequest(IPP_SET_JOB_ATTRIBUTES);
+  snprintf (uri, sizeof (uri), "ipp://localhost/jobs/%d", job_id);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI, "job-uri", NULL, uri);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+		"requesting-user-name", NULL, cupsUser ());
+  num_options = cupsAddOption ("job-hold-until", job_hold_until,
+			       num_options, &options);
+  cupsEncodeOptions (request, num_options, options);
+  debugprintf ("cupsDoRequest(\"/jobs/\")\n");
+  answer = cupsDoRequest (self->http, request, "/jobs/");
+  if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
+    set_ipp_error (answer ?
+		   answer->request.status.status_code :
+		   cupsLastError ());
+    if (answer)
+      ippDelete (answer);
+    debugprintf ("<- Connection_setJobHoldUntil() (error)\n");
+    return NULL;
+  }
+
+  Py_INCREF (Py_None);
+  debugprintf ("<- Connection_setJobHoldUntil() = None\n");
+  return Py_None;
+}
+
+static PyObject *
+Connection_restartJob (Connection *self, PyObject *args)
+{
+  ipp_t *request, *answer;
+  int job_id;
+  char uri[1024];
+  if (!PyArg_ParseTuple (args, "i", &job_id))
+    return NULL;
+
+  debugprintf ("-> Connection_restartJob(%d)\n", job_id);
+  request = ippNewRequest(IPP_RESTART_JOB);
+  snprintf (uri, sizeof (uri), "ipp://localhost/jobs/%d", job_id);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI, "job-uri", NULL, uri);
+  ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_NAME,
+		"requesting-user-name", NULL, cupsUser ());
+  debugprintf ("cupsDoRequest(\"/jobs/\")\n");
+  answer = cupsDoRequest (self->http, request, "/jobs/");
+  if (!answer || answer->request.status.status_code > IPP_OK_CONFLICT) {
+    set_ipp_error (answer ?
+		   answer->request.status.status_code :
+		   cupsLastError ());
+    if (answer)
+      ippDelete (answer);
+    debugprintf ("<- Connection_restartJob() (error)\n");
+    return NULL;
+  }
+
+  Py_INCREF (Py_None);
+  debugprintf ("<- Connection_restartJob() = None\n");
+  return Py_None;
+}
+
 static PyObject *
 Connection_getFile (Connection *self, PyObject *args)
 {
@@ -568,13 +853,17 @@ Connection_getFile (Connection *self, PyObject *args)
   if (!PyArg_ParseTuple (args, "ss", &resource, &filename))
     return NULL;
 
+  debugprintf ("-> Connection_getFile(%s, %s)\n", resource, filename);
+  debugprintf ("cupsGetFile()\n");
   status = cupsGetFile (self->http, resource, filename);
   if (status != HTTP_OK) {
     set_http_error (status);
+    debugprintf ("<- Connection_getFile() (error)\n");
     return NULL;
   }
 
   Py_INCREF (Py_None);
+  debugprintf ("<- Connection_getFile() = None\n");
   return Py_None;
 }
 
@@ -587,13 +876,17 @@ Connection_putFile (Connection *self, PyObject *args)
   if (!PyArg_ParseTuple (args, "ss", &resource, &filename))
     return NULL;
 
+  debugprintf ("-> Connection_putFile(%s, %s)\n", resource, filename);
+  debugprintf ("cupsPutFile()\n");
   status = cupsPutFile (self->http, resource, filename);
   if (status != HTTP_OK && status != HTTP_CREATED) {
     set_http_error (status);
+    debugprintf ("<- Connection_putFile() (error)\n");
     return NULL;
   }
 
   Py_INCREF (Py_None);
+  debugprintf ("<- Connection_putFile() = None\n");
   return Py_None;
 }
 
@@ -1260,9 +1553,9 @@ Connection_deletePrinterOptionDefault (Connection *self, PyObject *args)
 }
 
 static PyObject *
-Connection_deletePrinter (Connection *self, PyObject *args)
+Connection_deletePrinter (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, CUPS_DELETE_PRINTER);
+  return do_printer_request (self, args, kwds, CUPS_DELETE_PRINTER);
 }
 
 static PyObject *
@@ -1622,33 +1915,33 @@ Connection_deleteClass (Connection *self, PyObject *args)
 }
 
 static PyObject *
-Connection_enablePrinter (Connection *self, PyObject *args)
+Connection_enablePrinter (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, IPP_RESUME_PRINTER);
+  return do_printer_request (self, args, kwds, IPP_RESUME_PRINTER);
 }
 
 static PyObject *
-Connection_disablePrinter (Connection *self, PyObject *args)
+Connection_disablePrinter (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, IPP_PAUSE_PRINTER);
+  return do_printer_request (self, args, kwds, IPP_PAUSE_PRINTER);
 }
 
 static PyObject *
-Connection_acceptJobs (Connection *self, PyObject *args)
+Connection_acceptJobs (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, CUPS_ACCEPT_JOBS);
+  return do_printer_request (self, args, kwds, CUPS_ACCEPT_JOBS);
 }
 
 static PyObject *
-Connection_rejectJobs (Connection *self, PyObject *args)
+Connection_rejectJobs (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, CUPS_REJECT_JOBS);
+  return do_printer_request (self, args, kwds, CUPS_REJECT_JOBS);
 }
 
 static PyObject *
-Connection_setDefault (Connection *self, PyObject *args)
+Connection_setDefault (Connection *self, PyObject *args, PyObject *kwds)
 {
-  return do_printer_request (self, args, CUPS_SET_DEFAULT);
+  return do_printer_request (self, args, kwds, CUPS_SET_DEFAULT);
 }
 
 static PyObject *
@@ -1780,10 +2073,12 @@ Connection_adminSetServerSettings (Connection *self, PyObject *args)
   }
 
   if (!_cupsAdminSetServerSettings (self->http, num_settings, settings)) {
+    cupsFreeOptions (num_settings, settings);
     PyErr_SetString (PyExc_RuntimeError, "Failed to set settings");
     return NULL;
   }
 
+  cupsFreeOptions (num_settings, settings);
   Py_INCREF (Py_None);
   return Py_None;
 }
@@ -1816,6 +2111,24 @@ PyMethodDef Connection_methods[] =
       "Returns a dict, indexed by device URI, of dicts representing\n"
       "devices, indexed by attribute." },
 
+    { "getJobs",
+      (PyCFunction) Connection_getJobs, METH_VARARGS | METH_KEYWORDS,
+      "getJobs(which_jobs='not-completed', my_jobs=False) -> dict\n"
+      "Returns a dict, indexed by job ID, of dicts representing job\n"
+      "attributes." },
+
+    { "cancelJob",
+      (PyCFunction) Connection_cancelJob, METH_VARARGS,
+      "cancelJob(jobid) -> None" },
+
+    { "setJobHoldUntil",
+      (PyCFunction) Connection_setJobHoldUntil, METH_VARARGS,
+      "setJobHoldUntil(jobid, job_hold_until) -> None" },
+    
+    { "restartJob",
+      (PyCFunction) Connection_restartJob, METH_VARARGS,
+      "restartJob(jobid) -> None" },
+
     { "getFile",
       (PyCFunction) Connection_getFile, METH_VARARGS,
       "getFile(resource, filename) -> None" },
@@ -1890,8 +2203,8 @@ PyMethodDef Connection_methods[] =
       "CUPS 1.2.\n" },
 
     { "deletePrinter",
-      (PyCFunction) Connection_deletePrinter, METH_VARARGS,
-      "deletePrinter(name, ppdfile, device_uri) -> None" },
+      (PyCFunction) Connection_deletePrinter, METH_VARARGS | METH_KEYWORDS,
+      "deletePrinter(name) -> None" },
 
     { "getPrinterAttributes",
       (PyCFunction) Connection_getPrinterAttributes, METH_VARARGS,
@@ -1921,7 +2234,7 @@ PyMethodDef Connection_methods[] =
       "deleteClass(class) -> None" },
 
     { "setDefault",
-      (PyCFunction) Connection_setDefault, METH_VARARGS,
+      (PyCFunction) Connection_setDefault, METH_VARARGS | METH_KEYWORDS,
       "setDefault(name) -> None" },
 
     { "getPPD",
@@ -1929,19 +2242,21 @@ PyMethodDef Connection_methods[] =
       "Returns PPD file name" },
 
     { "enablePrinter",
-      (PyCFunction) Connection_enablePrinter, METH_VARARGS,
+      (PyCFunction) Connection_enablePrinter, METH_VARARGS | METH_KEYWORDS,
       "Enables named printer." },
 
     { "disablePrinter",
-      (PyCFunction) Connection_disablePrinter, METH_VARARGS,
+      (PyCFunction) Connection_disablePrinter, METH_VARARGS | METH_KEYWORDS,
+      "disablePrinter(name[,reason]) -> None\n\n"
       "Disables named printer." },
 
     { "acceptJobs",
-      (PyCFunction) Connection_acceptJobs, METH_VARARGS,
+      (PyCFunction) Connection_acceptJobs, METH_VARARGS | METH_KEYWORDS,
       "Causes named printer to accept jobs." },
 
     { "rejectJobs",
-      (PyCFunction) Connection_rejectJobs, METH_VARARGS,
+      (PyCFunction) Connection_rejectJobs, METH_VARARGS | METH_KEYWORDS,
+      "rejectJobs(name[,reason])\n\n"
       "Causes named printer to reject jobs." },
 
     { "printTestPage",
