@@ -1,6 +1,6 @@
 /*
  * cups - Python bindings for CUPS
- * Copyright (C) 2002, 2005, 2006  Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2002, 2005, 2006, 2007  Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,6 +21,7 @@
 #include "cupsmodule.h"
 
 #include <iconv.h>
+#include <stdbool.h>
 
 typedef struct
 {
@@ -63,20 +64,20 @@ ppd_encoding_is_utf8 (PPD *ppd)
     return 0;
 
   lang_encoding = ppd->ppd->lang_encoding;
-  if (!strcasecmp (lang_encoding, "UTF-8"))
+  if (lang_encoding && !strcasecmp (lang_encoding, "UTF-8"))
     return 1;
 
-  if (!strcasecmp (lang_encoding, "ISOLatin1"))
+  if (lang_encoding && !strcasecmp (lang_encoding, "ISOLatin1"))
     from_encoding = "ISO-8859-1";
-  else if (!strcasecmp (lang_encoding, "ISOLatin2"))
+  else if (lang_encoding && !strcasecmp (lang_encoding, "ISOLatin2"))
     from_encoding = "ISO-8859-2";
-  else if (!strcasecmp (lang_encoding, "ISOLatin5"))
+  else if (lang_encoding && !strcasecmp (lang_encoding, "ISOLatin5"))
     from_encoding = "ISO-8859-5";
-  else if (!strcasecmp (lang_encoding, "JIS83-RKSJ"))
+  else if (lang_encoding && !strcasecmp (lang_encoding, "JIS83-RKSJ"))
     from_encoding = "SHIFT-JIS";
-  else if (!strcasecmp (lang_encoding, "MacStandard"))
+  else if (lang_encoding && !strcasecmp (lang_encoding, "MacStandard"))
     from_encoding = "MACINTOSH";
-  else if (!strcasecmp (lang_encoding, "WindowsANSI"))
+  else if (lang_encoding && !strcasecmp (lang_encoding, "WindowsANSI"))
     from_encoding = "WINDOWS-1252";
   else
     // Guess
@@ -222,6 +223,14 @@ PPD_dealloc (PPD *self)
 // PPD // METHODS
 /////////
 
+static PyObject *
+PPD_localize (PPD *self)
+{
+  if (!ppdLocalize (self->ppd))
+    return Py_None;
+  return PyErr_SetFromErrno (PyExc_RuntimeError);
+}
+
 static PyObject *
 PPD_markDefaults (PPD *self)
 {
@@ -577,6 +586,10 @@ PyGetSetDef PPD_getseters[] =
 
 PyMethodDef PPD_methods[] =
   {
+    { "localize",
+      (PyCFunction) PPD_localize, METH_NOARGS,
+      "Localize PPD." },
+
     { "markDefaults",
       (PyCFunction) PPD_markDefaults, METH_NOARGS,
       "Mark default options.  Returns number of conflicts." },
@@ -746,6 +759,7 @@ Option_getChoices (Option *self, void *closure)
 {
   PyObject *choices = PyList_New (0);
   ppd_choice_t *choice;
+  bool defchoice_seen = false;
   int i;
 
   if (!self->option)
@@ -762,8 +776,24 @@ Option_getChoices (Option *self, void *closure)
 			  make_PyUnicode_from_ppd_string (self->ppd,
 							  choice->text));
     PyList_Append (choices, choice_dict);
+    if (!strcmp (choice->choice, self->option->defchoice))
+      defchoice_seen = true;
   }
 
+  if (!defchoice_seen) {
+    // This PPD option has a default choice that isn't one of the choices.
+    // This really happens.
+    const char *defchoice = self->option->defchoice;
+    PyObject *choice_dict = PyDict_New ();
+    PyDict_SetItemString (choice_dict, "choice",
+			  make_PyUnicode_from_ppd_string (self->ppd,
+							    defchoice));
+      PyDict_SetItemString (choice_dict, "text",
+			    make_PyUnicode_from_ppd_string (self->ppd,
+							    defchoice));
+      PyList_Append (choices, choice_dict);
+    }
+
   return choices;
 }
 
